import time
import heapq
import os
import inspect
import threading
import logging
import xxhash
from contextlib import contextmanager
from .diff import diff_dicts
from .utils import timing

logger = logging.getLogger("hyperdiv")
logger.setLevel(logging.DEBUG)

source_path = os.path.abspath(str(inspect.getsourcefile(lambda: 0)))


class EventBus:
    def __init__(self):
        self.listeners = dict()

    def on(self, event_name, listener):
        if event_name not in self.listeners:
            self.listeners[event_name] = []
        self.listeners[event_name].append(listener)

    def off(self, event_name, listener=None):
        if listener:
            if event_name in self.listeners:
                self.listeners[event_name].remove(listener)
                if len(self.listeners[event_name]) == 0:
                    self.listeners.pop(event_name)
        else:
            self.listeners.pop(event_name, None)

    def dispatch(self, event_name, *args, **kwargs):
        for listener in self.listeners.get(event_name, []):
            listener(*args, **kwargs)


class CollectorStack:
    def __init__(self):
        self.paused = False
        self.stack = []

    def push(self, collector):
        self.stack.append(collector)
        return collector

    def pop(self):
        return self.stack.pop()

    def current(self):
        return self.stack[-1]

    def collect(self, element):
        if not self.paused:
            self.current().append(element)


class ScopeStack:
    def __init__(self):
        self.stack = []

    def push(self, piece):
        self.stack.append(piece)

    def pop(self):
        return self.stack.pop()

    def __str__(self):
        return ":".join(self.stack)


class StateEntry:
    def __init__(self, element):
        self.element = element
        self.access_timestamp = None
        self.update_timestamp()

    def update_timestamp(self):
        self.access_timestamp = time.time()


class ElementState:
    # GC elements that haven't been accessed in at least `GC_MAX_AGE`
    # seconds
    GC_MAX_AGE = 5
    # GC every `GC_INTERVAL` runs
    GC_INTERVAL = 50

    def __init__(self):
        self.last_run_time = time.time()
        self.runs_since_last_gc = 0
        self.state = dict()
        self.lock = threading.RLock()

    def set_element(self, element):
        with self.lock:
            self.state[element.key] = StateEntry(element)

    def get_element(self, key):
        with self.lock:
            entry = self.state.get(key)
            if not entry:
                return None
            entry.update_timestamp()
            return entry.element

    def gc(self):
        """Called for every run."""
        self.last_run_time = time.time()
        self.runs_since_last_gc += 1

        if self.runs_since_last_gc < ElementState.GC_INTERVAL:
            return

        self.runs_since_last_gc = 0

        with self.lock:
            entries = sorted(
                list(self.state.values()),
                key=lambda entry: entry.access_timestamp,
            )

            for entry in entries:
                if entry.element.mutated():
                    continue
                if (
                    entry.access_timestamp
                    > self.last_run_time - ElementState.GC_MAX_AGE
                ):
                    # Since entries are sorted by access time,
                    # if we reach an entry that's too young to
                    # prune, all the ones after it will be
                    # younger.
                    break
                self.state.pop(entry.element.key)


class StateUpdateBatch:
    """Holds a batch of incoming state and location updates. The state
    updates have to be disjoint on key.
    """

    def __init__(self):
        self.state = dict()

    def has_state(self, key, prop_name):
        return (key, prop_name) in self.state

    def set_state(self, key, prop_name, prop_value):
        if self.has_state(key, prop_name):
            raise Exception(f"Duplicate update on {prop_name}.")
        self.state[(key, prop_name)] = prop_value


class PendingBatches:
    """A queue holding update batches. Context will fetch the next batch
    from this queue, apply it to its private state (by updating its
    ui_state and location), and doing one run per batch.

    A key property of this queue is that all the state updates in each
    batch are disjoint on key. If a batch B contains a key K, and
    there is an incoming update on K, K will not be added to
    B. Instead, a new batch will be created with K in it.

    This is because we want at least 1 run per key update. Otherwise,
    code like this will break:

        if button.clicked:
            count += 1

    If each key update doesn't get at last one run, that effectively
    amounts to dropping/forgetting events and incorrect behavior.
    """

    def __init__(self, event):
        # Event becomes set when a new batch arrives.
        self.event = event
        self.lock = threading.RLock()
        self.pending_batches = []

    def _add_batch(self):
        new_batch = StateUpdateBatch()
        self.pending_batches.append(new_batch)
        return new_batch

    def set_state_bulk(self, updates):
        """Add this all the (k, v) pairs in the given list to the first batch
        that does not contain any of the keys in updates, creating a
        batch if necessary.
        """

        if len(updates) == 0:
            return

        keys = set([(key, prop_name) for (key, prop_name, _) in updates])

        # Check for duplicates: Compare the lengths of the original
        # list and of a set of the (key, prop_name)s in the list,
        # which, being a set, implicitly drops the duplicates.
        if len(keys) != len(updates):
            raise Exception(
                "An id is not allowed to occur multiple times in bulk update."
            )

        with self.lock:
            target_batch = None
            for batch in self.pending_batches:
                is_good_batch = True
                for key, prop_name in keys:
                    if batch.has_state(key, prop_name):
                        is_good_batch = False
                        break
                if is_good_batch:
                    target_batch = batch
                    break

            if not target_batch:
                target_batch = self._add_batch()

            for key, prop_name, prop_value in updates:
                target_batch.set_state(key, prop_name, prop_value)
        self.event.set()

    def dequeue_batch(self):
        with self.lock:
            try:
                next_batch = self.pending_batches.pop(0)
            except IndexError:
                return None
            return next_batch


@contextmanager
def paused_collector():
    context = Context.current()
    context.collector_stack.paused = True
    try:
        yield
    finally:
        context.collector_stack.paused = False


class Context:
    """Every client connection gets one of these contexts. It wraps a ui
    state store and the collector stack that is used to collect dom
    during run().

    Whenever there is an update coming from the client, the client conneciton
    will update the ui state object and invoke run(), sending the updated dom to
    the client.
    """

    # The current context is stored here per thread.
    _local_store = threading.local()

    def __init__(self, main_function):
        self.main_function = main_function
        self.conn = None

        self.timing_enabled = False

        self.element_state = ElementState()
        # A collector stack is used to collect arbitrarily nested
        # elements.
        self.collector_stack = CollectorStack()
        # The ids seen during a run
        self.keys = set()
        # Current "scope"
        self.scope_stack = ScopeStack()

        self.prev_dom = None
        self.prev_listeners = set()

        # Lock used for thread-safe context methods.
        self.lock = threading.RLock()
        self.should_run = threading.Event()
        self.frame_had_changes = threading.Event()
        self.running_user_code = False

        self.stopped = threading.Event()
        self.pending_batches = PendingBatches(self.should_run)
        self.run_thread = None

        self.read_event_props = set()
        self.mutated_event_props = set()

        self.bus = EventBus()

        self.should_run.set()

        # Create the singletons, so they can be immediately updated
        # before a frame is run.

        Context.set_current_context(self)
        with paused_collector():
            self.call_singletons()
        Context.clear_current_context()

    def call_singletons(self):
        from .. import elements as el

        el.theme()
        el.head.head()
        el.location()

    def run_loop(self):
        def apply_update(key, prop_name, prop_value):
            element = self.element_state.get_element(key)
            if not element:
                logger.warning(
                    "Update attempted on unknown element. "
                    f"key={key} prop_name={prop_name}"
                )
                return
            if prop_name not in element._props:
                logger.warning(
                    "Update attempted on missing prop. "
                    f"key={key} prop_name={prop_name}"
                )
                return
            element._props[prop_name].set_value(prop_value)

        while True:
            if self.stopped.is_set():
                break

            if not self.conn.opened.wait(0.01):
                continue

            if not self.should_run.wait(0.01):
                continue

            self.should_run.clear()

            batch = self.pending_batches.dequeue_batch()
            if batch:
                for (key, prop_name), prop_value in batch.state.items():
                    apply_update(key, prop_name, prop_value)

            dom = self.run()
            if dom:
                self.conn.send(dom)

            self.element_state.gc()

        logger.info("Run loop exiting...")

    def start(self, conn):
        self.conn = conn
        self.run_thread = threading.Thread(target=self.run_loop)
        self.stopped.clear()
        self.run_thread.start()

    def stop(self):
        # TODO: Don't join on the ioloop.
        self.stopped.set()
        self.run_thread.join()
        self.run_thread = None
        self.conn = None

    def set_state_bulk(self, updates):
        """Thread-safe. This is how an external thread, like the ioloop,
        applies multiple updates atomically.
        """
        self.pending_batches.set_state_bulk(updates)

    def set_state(self, id, k, v):
        """Thread-safe. This is how an external thread, like the ioloop,
        applies a single state update.
        """
        self.pending_batches.set_state_bulk([(id, k, v)])

    def refresh(self):
        """Thread-safe. This is how code inside or outside the frame (like an
        async callback) signals that one more run should be done.
        """
        with self.lock:
            self.should_run.set()

    #
    # These static methods manage the "current context", of which
    # there can be only one per thread.
    #

    @staticmethod
    def _get_current_context():
        if hasattr(Context._local_store, "context") and Context._local_store.context:
            return Context._local_store.context
        else:
            return None

    @staticmethod
    def set_current_context(context):
        current_context = Context._get_current_context()
        if current_context:
            raise Exception("There is already an active context.")

        Context._local_store.context = context

    @staticmethod
    def clear_current_context():
        Context._local_store.context = None

    @staticmethod
    def current():
        context = Context._get_current_context()
        if not context:
            raise Exception("No active context found.")
        return context

    def extract_key_from_call_stack(self):
        frame = inspect.currentframe().f_back.f_back.f_back
        key = ""
        while frame:
            filename = frame.f_code.co_filename
            lineno = frame.f_lineno
            funcname = frame.f_code.co_name
            if filename == source_path and funcname == "_run":
                break
            key += f"::{filename}:{lineno}:{funcname}"
            frame = frame.f_back
        key = str(self.scope_stack) + key
        key = "a" + xxhash.xxh32(key).hexdigest()
        return key

    def create_element_key(self, key, duplicate_check=True):
        """Generate an ID from the source code position of the caller, by
        inspecting the stack. NOTE: This is brittle and may break as we shuffle
        code around.

        Assumptions:
        * The 0 frame is this function: get_id
        * The 1 frame is a hyperdiv library function or constructor in this
          module, like checkbox, button, etc.
        * The 2 frame is client code calling into hyperdiv.
        * ...
        * The Nth frame is the function that runs the client code. Currently we
          assume it's located in this source file and is named "_run".

        This code walks the stack starting at frame 2 and builds up an ID by
        appending the "filename:funcname:line_number" for each frame up until we
        encounter the _run function.

        This strategy allows clients to build UIs from modular functions. For
        example, stuff like this works:

            def element(text, label):
                hd.markdown(text)
                hd.input_text(label)
                hd.button("Add")

            def main():
                element("Foo", "Bar")
                element("Bar", "Baz")

            hd.run(main)

        For example `hd.markdown(text)` will get a different ID in the first
        call to `element()` than in the 2nd call, because ID generation takes
        into account the different source positions of the element() calls.
        """
        if not key:
            key = self.extract_key_from_call_stack()

        if duplicate_check:
            if key in self.keys:
                raise Exception(f"Duplicate Key: {key}")
            self.keys.add(key)
        return key

    def process_end_of_frame_event_props(self):
        for prop in self.mutated_event_props:
            prop.reset()

        listeners = set(
            [(prop.parent_element.key, prop.name) for prop in self.read_event_props]
        )

        self.mutated_event_props.clear()
        self.read_event_props.clear()

        return listeners

    def in_user_code(self):
        return threading.current_thread() == self.run_thread and self.running_user_code

    def run(self):
        """This is the meat of the framework. It sets this context to be current
        and runs the main function in it. Hyperdiv client code that looks up
        state and collects dom will use the ui state and collectors in this
        context.
        """
        Context.set_current_context(self)

        def _run():
            from .. import elements as el

            self.bus.dispatch("frame_start")

            # TODO: Maybe users should be able to set props on the
            # root container.

            with paused_collector():
                theme = el.theme()
                root_container = el.container(
                    key="root-container",
                    background_color=theme.background_color,
                    font_color=theme.font_color,
                    font_family=theme.font_family,
                    height="100%",
                    horizontal_scroll=True,
                    vertical_scroll=True,
                )

            with root_container:
                self.call_singletons()
                self.running_user_code = True
                self.main_function()
                self.running_user_code = False

            self.keys.clear()

            listeners = self.process_end_of_frame_event_props()
            return root_container, listeners

        try:
            # Do at least one run
            root_container, listeners = _run()
            run_count = 1

            # While the code keeps mutating props, keep running it
            # until there's a frame with no mutations, or we get to an
            # upper limit that may indicate an infinite loop.
            while True:
                if not self.frame_had_changes.is_set():
                    break
                self.frame_had_changes.clear()
                root_container, listeners = _run()
                run_count += 1
                if run_count >= 40:
                    raise Exception(
                        "Too many internal refreshes. Likely an infinite loop."
                    )

            dom = root_container.to_dom()

            #
            # Diff listeners
            #

            listeners_to_add = listeners.difference(self.prev_listeners)
            listeners_to_remove = self.prev_listeners.difference(listeners)
            self.prev_listeners = listeners

            return_value = dict()

            if listeners_to_add or listeners_to_remove:
                listeners_diff = dict()
                if listeners_to_add:
                    listeners_diff["add"] = list(listeners_to_add)
                if listeners_to_remove:
                    listeners_diff["remove"] = list(listeners_to_remove)
                return_value["listeners"] = listeners_diff

            #
            # Diff dom
            #

            if not self.prev_dom:
                # Send the full dom. This only happens on the first
                # request.
                self.prev_dom = dom
                return_value["dom"] = self.prev_dom
                return_value["clientId"] = self.conn.client_id.hex
            else:
                # Send a diff, or nothing if there's no difference.
                d = diff_dicts(self.prev_dom, dom)

                self.prev_dom = dom
                if len(d) > 0:
                    return_value["diff"] = d

            return return_value if len(return_value) > 0 else None
        finally:
            Context.clear_current_context()
