from nose.tools import assert_raises
import hyperdiv as hd
from .. import element, context
from ..unittests import simple_test, should_raise


def test_to_caml_case():
    assert element.to_caml_case("foo") == "foo"
    assert element.to_caml_case("foo_bar") == "fooBar"
    assert element.to_caml_case("foo_bar_baz") == "fooBarBaz"
    assert element.to_caml_case("") == ""


def test_parse_option():
    parse = element.parse_option(None, "hey", "there", "dude")

    assert parse("hey") == "hey"
    assert parse("there") == "there"
    assert parse("dude") == "dude"
    assert parse(None) is None

    assert_raises(Exception, parse, "stuff")


class MockElement:
    def __init__(self):
        self._context = context.Context(lambda: None)


def _prop(*args, **kwargs):
    return element.DynamicProp(MockElement(), *args, **kwargs)


def test_dynamic_prop():
    prop = _prop("my_prop", parse=element.parse_option("one", "two"))
    prop.init("one")
    assert prop.get_value() == "one"
    prop.init("two")
    assert prop.get_value() == "two"
    prop.set_value("one")
    assert prop.get_value() == "one"
    # Init after mutation has no effect
    prop.init("two")
    assert prop.get_value() == "one"
    # Init works again after clearing
    prop.reset()
    prop.init("two")
    assert prop.get_value() == "two"

    assert_raises(Exception, prop.init, "blah")


@simple_test
def test_element_cache(ctx):
    singleton_keys = ["theme", "location", "head"]
    assert set(ctx.element_state.state.keys()) == set(singleton_keys)
    el = hd.button()
    assert set(ctx.element_state.state.keys()) == set(singleton_keys + [el.key])
    assert ctx.element_state.state[el.key] == el
    el1 = hd.text()
    assert set(ctx.element_state.state.keys()) == set(
        singleton_keys + [el.key, el1.key]
    )
    assert ctx.element_state.state[el1.key] == el1


def test_collect_static_props():
    class Foo(element.Element):
        a = element.Prop("0")
        b = element.Prop("1")

    assert element.collect_static_props(Foo) == dict(a=Foo.a, b=Foo.b)

    class Bar(Foo):
        a = element.Prop("2")
        c = element.Prop("3")

    assert element.collect_static_props(Bar) == dict(a=Bar.a, c=Bar.c, b=Foo.b)

    class Baz(Bar):
        a = element.RemoveProp()
        c = element.Prop("4")
        d = element.Prop("5")
        x = 1

    assert element.collect_static_props(Baz) == dict(b=Foo.b, c=Baz.c, d=Baz.d)


def test_prop_accessor():
    acc = element.PropAccessor(dict(x=_prop("x", 1), y=_prop("y", 2), z=_prop("z", 3)))
    assert acc.x == 1
    assert acc.y == 2
    assert acc.z == 3

    acc.x = 100
    assert acc._props["x"].get_value() == 100
    assert acc.x == 100

    # TODO: It would be great if assignment could create a prop if it
    # does not exist.
    acc.p = 100
    assert acc.p == 100
    assert "p" not in acc._props


@simple_test
def test_callable_prop_value(ctx):
    class elem(element.Element):
        a = element.Prop(10)
        b = element.Prop(lambda el: el.a * 3)
        c = element.Prop()

    foo = elem()
    assert foo.b == 30
    foo.a = 11
    assert foo.b == 33

    foo.b = 100
    assert foo.b == 100

    foo.a = 3
    foo.b = lambda el: el.a * 100
    assert foo.b == 300

    class singleton_elem(element.Element):
        is_singleton = True
        blah = "hello"

    singleton_elem()
    foo.c = lambda el: singleton_elem().blah + " world"
    assert foo.c == "hello world"


@simple_test
def test_element_basics(ctx):
    class elem(element.Element):
        x = element.Prop(1)

    e = elem(x=2)
    assert e.x == 2

    # Cannot init prop 'y'
    assert_raises(Exception, elem, x=2, y=3)

    # Cannot enter a non-container:
    with should_raise():
        with elem():
            pass


@simple_test
def test_element_container(ctx):
    class elem(element.Element):
        is_container = True
        x = element.Prop(1)

    with elem(x=2) as e:
        t = hd.plaintext("hello")

    assert e.x == 2
    assert e.children == [t]


@simple_test
def test_element_singleton(ctx):
    class elem(element.Element):
        is_singleton = True

    e1 = elem()
    e2 = elem()

    assert e1.key == e2.key


@simple_test
def test_element_to_dom(ctx):
    class elem(element.Element):
        is_container = True
        x = element.Prop(1)
        y = element.Prop(2)

    e = elem()
    assert e.to_dom() == dict(_dom=True, name="elem", key=e.key, x=1, y=2, children=[])

    class elem1(elem):
        y = element.Prop(3)
        z = element.Prop(20)

    e = elem1()
    with e:
        t = hd.plaintext("hello")

    assert e.to_dom() == dict(
        _dom=True,
        name="elem1",
        key=e.key,
        x=1,
        y=3,
        z=20,
        children=[
            dict(
                _dom=True,
                name="plaintext",
                key=t.key,
                label="hello",
            )
        ],
    )
