from ..core.element import Prop, parse_option
from .styled_element import StyledElement
from .style_utils import parse_size, parse_non_negative_number, is_fr


alignments = ("start", "end", "center", "stretch")
directions = ("vertical", "horizontal")
justifies = (
    "start",
    "end",
    "center",
    "space-between",
    "space-around",
    "space-evenly",
    "stretch",
)
wraps = ("nowrap", "wrap", "reverse")


def parse_wrap(wrap):
    if wrap is True:
        return "wrap"
    if wrap is False:
        return "nowrap"
    if wrap in wraps:
        return wrap
    raise ValueError("Invalid wrap value.")


def parse_aria(aria):
    if aria is None:
        return None
    if not isinstance(aria, dict):
        raise ValueError("aria must be a dict.")
    for k, v in aria.items():
        if not (isinstance(k, str) and k.isalpha()):
            raise ValueError("aria keys must be alphabetic strings.")
        if not isinstance(v, (str, int, float, bool)):
            raise ValueError("aria values must be primitive.")
    return aria


def parse_gap(gap):
    if type(gap) not in (list, tuple):
        return parse_size(gap)

    else:
        return f"{parse_size(gap[0])} {parse_size(gap[1])}"


def parse_grid_row(gridrow):
    if gridrow is None:
        return None
    if type(gridrow) in (list, tuple) and len(gridrow) == 2:
        return f"{gridrow[0]} / {gridrow[1]}"

    raise ValueError("Invalid grid row / column property. Must be a tuple of length 2.")


def parse_grid_template_areas(areas):
    if areas is None:
        return None
    elif type(areas) in (tuple, list):
        return areas

    raise ValueError("Invalid grid_template_areas format.")


def parse_grid_template_rows(rows):
    if rows is None:
        return None
    elif type(rows) in (tuple, list):
        rowstr = ""
        for i in rows:
            if type(i) is int:
                i = parse_non_negative_number(i)
            elif i != "auto" and not is_fr(i):
                raise ValueError("Invalid grid_template_rows / columns.")
            rowstr += f"{i} "

        return rowstr

    raise ValueError("Invalid grid_template_rows / columns.")


class container(StyledElement):
    is_container = True

    direction = Prop("vertical", parse=parse_option(*directions))
    vertical_scroll = Prop(False)
    horizontal_scroll = Prop(False)
    align = Prop("stretch", parse=parse_option(*alignments))
    justify = Prop("start", parse=parse_option(*justifies))
    gap = Prop(parse=parse_gap)
    layout = Prop("flex", parse=parse_option("flex", "grid"))
    grid_area = Prop()
    wrap = Prop(False, parse=parse_wrap)
    role = Prop("")
    aria = Prop(parse=parse_aria)
    grid_template_columns = Prop(parse=parse_grid_template_rows)
    grid_template_rows = Prop(parse=parse_grid_template_rows)
    grid_template_areas = Prop(parse=parse_grid_template_areas)
    grid_row = Prop(parse=parse_grid_row)
    grid_column = Prop(parse=parse_grid_row)
