from nose.tools import assert_raises
from ...core.unittests import simple_test
from ..style_utils import (
    parse_non_negative_number,
    parse_border_radius,
    parse_color,
    parse_cursor,
    parse_height,
    parse_padding,
    parse_border,
    parse_width,
    parse_font_family,
    parse_font_weight,
    parse_font_size,
)


def test_parse_font_size():
    assert parse_font_size("medium") == "medium"
    assert_raises(ValueError, parse_font_size, 10)


def test_parse_font_weight():
    assert parse_font_weight("bold") == "bold"
    assert_raises(ValueError, parse_font_weight, "blue")


def test_parse_font_family():
    assert parse_font_family(None) is None
    assert parse_font_family(("helvetica", "sans-serif")) == "helvetica, sans-serif"
    assert parse_font_family("helvetica") == "helvetica"


def test_parse_non_negative_number():
    assert parse_non_negative_number(0) == 0
    assert parse_non_negative_number(1.3) == 1.3
    assert_raises(ValueError, parse_non_negative_number, -1)
    assert_raises(ValueError, parse_non_negative_number, "hello")


def test_parse_padding():
    assert parse_padding(1) == 1
    assert parse_padding([1]) == dict(top=1, bottom=0, left=0, right=0)
    assert parse_padding([1, 2]) == dict(top=1, bottom=0, left=0, right=2)
    assert parse_padding([1, 2, 3]) == dict(top=1, bottom=3, left=0, right=2)
    assert parse_padding([1, 2, 3, 4]) == dict(top=1, bottom=3, left=4, right=2)
    assert parse_padding(dict(top=3, bottom=2)) == dict(
        top=3, bottom=2, left=0, right=0
    )
    assert_raises(ValueError, parse_padding, "Hello")
    assert_raises(ValueError, parse_padding, ["Bunnies", "Squirrels"])
    assert_raises(ValueError, parse_padding, dict(random_stuff="Bunnies"))
    assert_raises(ValueError, parse_padding, -1)


@simple_test
def test_parse_border(ctx):

    border = (1, "solid", "#eee")
    border_str = "1px solid #eee"
    assert parse_border(border) == border_str
    assert parse_border(None) is None
    assert parse_border([border]) == dict(
        top=border_str, bottom=None, left=None, right=None
    )
    assert parse_border([border, border]) == dict(
        top=border_str, bottom=None, left=None, right=border_str
    )
    assert parse_border([border, None, border]) == dict(
        top=border_str, bottom=border_str, left=None, right=None
    )
    assert parse_border([border, None, border, border]) == dict(
        top=border_str, bottom=border_str, left=border_str, right=None
    )

    assert_raises(ValueError, parse_border, "Stuff")
    assert_raises(
        ValueError,
        parse_border,
        [border, border, border, border, border, border],
    )
    assert_raises(ValueError, parse_border, [])
    assert_raises(ValueError, parse_border, ["Hello", object()])


def test_parse_border_radius():
    assert parse_border_radius(1) == 1
    assert parse_border_radius([1]) == dict(
        topleft=1, topright=0, bottomright=0, bottomleft=0
    )
    assert parse_border_radius([1, 2]) == dict(
        topleft=1,
        topright=2,
        bottomright=0,
        bottomleft=0,
    )
    assert parse_border_radius([1, 2, 3]) == dict(
        topleft=1, topright=2, bottomright=3, bottomleft=0
    )
    assert parse_border_radius([1, 2, 3, 4]) == dict(
        topleft=1, topright=2, bottomright=3, bottomleft=4
    )

    assert_raises(ValueError, parse_border_radius, "Hello")
    assert_raises(ValueError, parse_border_radius, ["Bunnies", "Squirrels"])
    assert_raises(ValueError, parse_border_radius, dict(random_stuff="Bunnies"))
    assert_raises(ValueError, parse_border_radius, -1)


def test_parse_width():
    assert parse_width(1) == "1px"
    assert parse_width("50%") == "50%"
    assert parse_width("10vw") == "10vw"
    assert parse_width(None) is None
    assert_raises(ValueError, parse_width, "Hello")
    assert_raises(ValueError, parse_width, -1)
    assert_raises(ValueError, parse_width, dict(width="50px"))


def test_parse_height():
    assert parse_height(1) == "1px"
    assert parse_height("50%") == "50%"
    assert parse_height("10vh") == "10vh"
    assert parse_height(None) is None
    assert_raises(ValueError, parse_height, "Hello")
    assert_raises(ValueError, parse_height, -1)
    assert_raises(ValueError, parse_height, dict(width="50px"))


def test_parse_cursor():
    assert parse_cursor("auto") == "auto"
    assert parse_cursor("pointer") == "pointer"
    assert parse_cursor(None) is None
    assert_raises(ValueError, parse_cursor, "hello")
    assert_raises(ValueError, parse_cursor, 1)


@simple_test
def test_parse_color(ctx):
    import hyperdiv as hd

    assert parse_color("#fff") == "#fff"
    assert parse_color("#fffeee") == "#fffeee"
    assert parse_color("#fffeeeaa") == "#fffeeeaa"
    assert parse_color("red") == hd.theme().red
    assert parse_color(None) is None

    assert_raises(ValueError, parse_color, "#ffeefff")
    assert_raises(ValueError, parse_color, "ffeeff")
    assert_raises(ValueError, parse_color, "#ffzzff")
    assert_raises(ValueError, parse_color, "hello")
    assert_raises(ValueError, parse_color, 1)
